/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#ifndef WIN32
# include <sys/stat.h>
# include <pwd.h>
#endif /* WIN32 */

#include <QMainWindow>
#include <QMessageBox>
#include <QLabel>
#include <QDir>
#include <QCloseEvent>
#include <QSettings>
#include <QFont>
#include <QFileInfo>
#include <QTimer>
#include <QUrl>
#include <QDesktopServices>

#include "mainwindow.h"
#include "puzzle.h"
#include "settings.h"
#include "settingspersist.h"
#include "help.h"

#include "../logic/defines.h"
#include "../logic/dict.h"
#include "../logic/util.h"
#include "../logic/game.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/defines.h"

using std::string;

const 	int MAX_PATH_LEN = 1024;
const	char STATEFLMAGIC[] = "sfm";
char 	stateFlPath[MAX_PATH_LEN] = "";
vector<string> setWords;

struct gameStateHeader {
	char	magic[4];
	int		time;
	char	game[MAX_LETRS+1];
	int 	hint;
	int		nwords;
} h;

struct Widgets {
	Puzzle		*puzzle;
	QGroupBox	*puzzleBox;

	Settings	*settings;

	QStatusBar	*statusOutput;
	QLabel		*statusWordCnt;
	QLabel		*statusMsg;

	QPushButton *play;
	QPushButton *hint;
	QPushButton *finish;
	QPushButton *solve;

	QTextEdit	*wordsOutput;
	WordInput	*wordInput;
	Help	*help;
};

struct Actions {
	QAction		*help;
	QAction		*about;
	QAction		*settings;
	QAction		*website;
	QAction		*exit;
	QAction		*aboutQt;
	QAction		*loadGame;
	QAction		*saveGame;
};

struct State {
	QString		dictChanged;
	bool		helpOpened;
	bool		ingame;
	bool		inprogress;	/* game started and atleast one correct word \
							   entered */
};

struct MainWindow::Private {
	Private(MainWindow *m);
	~Private();

	void createUi();
	void createConnections();
	void applySettings();

	MainWindow	*p;
	Widgets		widgets;
	Actions		actions;
	State		state;

	SettingsPersist	settings;
	QTimer		*timer;
	Spellathon::Dict		*dict;
	Spellathon::Game		*game;
	Spellathon::DispWordsBuf *dsp;
	time_t		t;
};

MainWindow::Private::Private(MainWindow *m)
	: p(m), timer(0), dict(0), game(0), dsp(0)
{
	state.dictChanged	= "";
	state.helpOpened	= false;
	state.ingame		= false;
	state.inprogress	= false;
}

MainWindow::Private::~Private()
{
	settings.save();

	delete dsp;
	delete game;
	delete dict;
}

void
MainWindow::Private::createUi()
{
	p->setupUi(p);

	widgets.puzzleBox		= p->puzzleBox;
	widgets.statusOutput	= p->statusOutput;
	widgets.hint			= p->hintButton;
	widgets.play			= p->playButton;
	widgets.finish			= p->finishButton;
	widgets.solve			= p->solveButton;
	widgets.wordsOutput		= p->wordsOutput;
	widgets.wordInput		= p->wordInput;

	widgets.help			= new Help(p);
	widgets.puzzle			= new Puzzle(widgets.puzzleBox);
	widgets.settings		= new Settings(p, &settings);
	widgets.statusWordCnt	= new QLabel(p);
	widgets.statusMsg		= new QLabel(p);

	actions.help			= p->action_Instructions;
	actions.about			= p->action_About;
	actions.settings		= p->action_Settings;
	actions.website			= p->action_Website;
	actions.exit			= p->action_Exit;
	actions.aboutQt			= p->action_About_Qt;
	actions.loadGame		= p->action_LoadGame;
	actions.saveGame		= p->action_SaveGame;

	widgets.statusWordCnt->setText("          ");
	widgets.statusOutput->insertWidget(0, widgets.statusMsg, 1);
	widgets.statusOutput->insertPermanentWidget(1, widgets.statusWordCnt, 0);
	widgets.hint->setText("H&int");
	widgets.hint->setEnabled(false);
	widgets.finish->setEnabled(false);
	widgets.wordInput->setEnabled(false);
	QHBoxLayout *layout = new QHBoxLayout;
	layout->setAlignment(Qt::AlignVCenter|Qt::AlignHCenter);
	layout->addWidget(widgets.puzzle);
	widgets.puzzleBox->setLayout(layout);
	p->setWindowIcon(QIcon(":/spth.ico"));

	timer = new QTimer(p);
	timer->setInterval(1000);

	dict = new Spellathon::Dict(p);
	game = new Spellathon::Game(dict, settings.pluralSieve);
	dsp = new Spellathon::DispWordsBuf;

	createConnections();
}

bool
MainWindow::initBackend()
{
	try {
		d->dict->init(d->settings.dictPath.toAscii().data(), 
					  d->settings.dictPrecomp, d->settings.pluralSieve);
	}
	catch (string s) 
	{
		QMessageBox::critical(this, "Error", s.c_str());
		disableSpellathon();
		return false;
	}
	
	d->widgets.statusOutput->showMessage("Dictionary loaded.");
	enableSpellathon();
	if (d->settings.savedGame) {
		if (savedGameExists()) 
			play();
		else 
			d->actions.loadGame->setEnabled(false);
		d->settings.savedGame = false;
	}
	return true;
}

void
MainWindow::Private::applySettings()
{

	p->resize(settings.mainwindowSize);
	p->move(settings.mainwindowPos);

	QFont monoFont = settings.monoFont;
	widgets.wordsOutput->setFont(monoFont);
	QFont f = monoFont;
	f.setBold(true);
	widgets.puzzle->setFont(f);
}

void
MainWindow::Private::createConnections()
{
	connect(timer, SIGNAL(timeout()), 
			p, SLOT(updateTime()));
	connect(actions.help, SIGNAL(triggered()), 
			widgets.help, SLOT(show()));
	connect(actions.about, SIGNAL(triggered()), 
			p, SLOT(about()));
	connect(actions.settings, SIGNAL(triggered()),
			widgets.settings, SLOT(show()));
	connect(actions.website, SIGNAL(triggered()),
			p, SLOT(openWWW()));
	connect(actions.exit, SIGNAL(triggered()),
			p, SLOT(close()));
	connect(actions.aboutQt, SIGNAL(triggered()),
			p, SLOT(aboutQt()));
	connect(actions.loadGame, SIGNAL(triggered()),
			p, SLOT(loadGame()));
	connect(actions.saveGame, SIGNAL(triggered()),
			p, SLOT(saveGameSlot()));

	connect(widgets.play, SIGNAL(clicked()), 
			p, SLOT(play()));
	connect(widgets.hint, SIGNAL(clicked()), 
			widgets.wordInput, SLOT(setFocus()));
	connect(widgets.hint, SIGNAL(clicked()), 
			p, SLOT(hint()));
	connect(widgets.finish, SIGNAL(clicked()), 
			p, SLOT(finish()));
	connect(widgets.solve, SIGNAL(clicked()), 
			p, SLOT(solve()));

	connect(widgets.wordInput->wordEdit, SIGNAL(returnPressed()), 
			p, SLOT(wordEntered()));
	connect(widgets.wordsOutput, SIGNAL(selectionChanged()),
			widgets.wordsOutput, SLOT(copy()));
	connect(widgets.settings, SIGNAL(settingsAccepted()),
			p, SLOT(settingsAccepted()));
}

MainWindow::MainWindow(QWidget *parent)
	: QMainWindow(parent), d(new MainWindow::Private(this))
{
	d->settings.load();
	d->createUi();
	d->applySettings();
	/* idle processing for initBackend() */
	QTimer::singleShot(0, this, SLOT(initBackend()));
}

void
MainWindow::settingsAccepted()
{
		if (d->settings.pluralSieve != d->dict->pluralSieve() ||
			d->settings.dictPath != d->dict->dictPath().c_str() ||
			d->settings.dictPrecomp != d->dict->precomp()) {

			if (d->state.ingame) {
				d->state.dictChanged = d->settings.dictPath;
				d->widgets.statusOutput->showMessage("Dictionary change pending...", 3000);
			} else {
				initBackend();
			}
		}

		if (d->settings.pluralSieve != d->game->pluralSieve())
				d->game->setPluralSieve(d->settings.pluralSieve);

		if (d->settings.monoFont != d->widgets.wordsOutput->font()) {
			d->widgets.wordsOutput->setFont(d->settings.monoFont); 
			QFont f = d->settings.monoFont;
			f.setBold(true);
			d->widgets.puzzle->setFont(f);
			wpl_update();
		}
}

void 
MainWindow::about()
{
	const char* svn_version();
	QMessageBox::about(this, "About", QString("<font size=\"+2\">Spellathon </font>"VERSION"&nbsp;&nbsp;<font color=\"gray\">(svn%1)</font><br><br>&copy; 2009 Girish Rao<br>").arg(svn_version()));
}

void 
MainWindow::aboutQt()
{
	QMessageBox::aboutQt(this, "About Qt");
}

bool
MainWindow::savedGameExists()
{
	bool r = true;
	try {
		setStateFlPath();
	}
	catch (string s) {
		QMessageBox::critical(this, "Error", s.c_str());
	}

	FILE	*fl;
	if ((fl = fopen(stateFlPath, "rb")) == NULL) 
		r = false;
	else
		fclose(fl);

	return r;
}

void 
MainWindow::loadGame()
{
		d->settings.savedGame = true;
		play();
}

void 
MainWindow::saveGame()
{
	if (d->state.ingame) {
		int n = d->dsp->setWords().size();

		h.time = d->t;
		strcpy(h.game, d->game->puzzle().c_str());
		strcpy(h.magic, STATEFLMAGIC);
		h.nwords = n;
		h.hint = d->dsp->hintLevel();

		FILE *fl = NULL;
		try {
			setStateFlPath();
			fl = fopen(stateFlPath, "wb");
			if (fl == NULL)
 				       Spellathon::spth_error(1, "Cannot open '%s'", stateFlPath);
			if (fwrite(&h, sizeof(gameStateHeader), 1, fl) != 1)
 				       Spellathon::spth_error(1, "'%s': Write error", stateFlPath);
			for (int i=0; i < n; i++) {
				char w[MAX_LETRS+1];
				strncpy(w, (d->dsp->setWords())[i].c_str(), MAX_LETRS+1);
				if (fwrite(w, MAX_LETRS+1, 1, fl) != 1)
 				       Spellathon::spth_error(1, "'%s': Write error", stateFlPath);
			}
			fclose(fl);
		}
		catch (string s) {
			QMessageBox::critical(this, "Error", s.c_str());
			if (fl != NULL)
				fclose(fl);
			d->settings.savedGame = false;
			return;
		}
		d->actions.loadGame->setEnabled(true);
		d->widgets.statusOutput->showMessage("Game saved.", 3000);
	} 
}

void 
MainWindow::saveGameSlot()
{
	if (!d->state.ingame) 
		QMessageBox::information(this, "Save game", "No game running, nothing to save.");
	else 
		saveGame();
}

void 
MainWindow::enableHints(int n)
{
	switch (n) {
	case -1:
			d->widgets.hint->setText("H&int");
			d->widgets.hint->setEnabled(false);
			break;
	case 0:
			d->widgets.hint->setText("H&int (0)");
			d->widgets.hint->setEnabled(false);
			break;
	case 1:
			d->widgets.hint->setText("H&int (1)");
			d->widgets.hint->setEnabled(true);
			break;
	case 2:
			d->widgets.hint->setText("H&int (2)");
			d->widgets.hint->setEnabled(true);
			break;
	}
}

void
MainWindow::enableSpellathon()
{
	d->widgets.play->setEnabled(true);
	d->widgets.play->setFocus();
	d->widgets.solve->setEnabled(true);
	d->widgets.puzzle->setReadOnly(false);
}

void
MainWindow::disableSpellathon()
{
	d->timer->stop();
	d->widgets.play->setEnabled(false);
	enableHints(-1);
	d->widgets.solve->setEnabled(false);
	d->widgets.finish->setEnabled(false);
	d->widgets.wordInput->setEnabled(false);
	d->widgets.puzzle->clear();
	d->widgets.puzzle->setReadOnly(true);
	d->widgets.wordInput->wordEdit->setInitState(false);
	d->widgets.wordsOutput->clear();
	d->widgets.statusMsg->clear();
	d->widgets.statusWordCnt->setText("          ");
}

void 
MainWindow::play()
{
	d->widgets.statusOutput->clearMessage();
	if (d->state.inprogress) {
		int r = QMessageBox::warning(this, "New Game", 
					     	"Game in progress. Quit current game?", 
					     	QMessageBox::Cancel | QMessageBox::Ok, 
							QMessageBox::Cancel);

		if (r == QMessageBox::Cancel)
			return;
	}

	d->state.ingame = false;
	d->state.inprogress = false;

	if (d->state.dictChanged != "") {
		if (!initBackend())
			return;
		d->widgets.statusOutput->showMessage("New dictionary loaded.", 3000);
		d->state.dictChanged = "";
	}

	d->timer->stop();

	bool r;

	if (d->settings.savedGame) {
		FILE *fl = NULL;
		try {
			if (stateFlPath[0] == 0)
				setStateFlPath();

			fl = fopen(stateFlPath, "rb");
			if (fl == NULL)
 			       Spellathon::spth_error(1, "Cannot open '%s'", stateFlPath);
			if (fread(&h, sizeof(h), 1, fl) != 1)
 				       Spellathon::spth_error(1, "'%s': Read error", stateFlPath);

			if (strcmp(h.magic, STATEFLMAGIC) != 0)
 			       Spellathon::spth_error(1, "'%s': Saved game file is corrupt.", stateFlPath);

			setWords.clear();

			for (int i=0; i < h.nwords; i++) {
				char word[MAX_LETRS+1];
				if (fread(word, MAX_LETRS+1, 1, fl) != 1)
 				       Spellathon::spth_error(1, "'%s': Read error", stateFlPath);

				setWords.push_back(word);
			}
			fclose(fl);
		}
		catch (string s) {
			QMessageBox::critical(this, "Error", s.c_str());
			d->settings.savedGame = false;
			if (fl != NULL)
				fclose(fl);
			return;
		}

		r = d->game->play(d->settings.dictPrecomp, h.game);
	} else {
		// this may take some time so grab mouse
		grabMouse(QCursor(Qt::WaitCursor));
		r = d->game->play(d->settings.dictPrecomp, d->settings.minSol, 
						d->settings.maxSol);
		releaseMouse();
	}


	if (r) {
		d->state.ingame = true;
		enableHints(2);
		d->widgets.finish->setEnabled(true);
		d->widgets.solve->setEnabled(false);
		d->widgets.wordInput->setEnabled(true);
		d->widgets.wordInput->setFocus();
		d->widgets.wordInput->wordEdit->setInitState(true);

		d->widgets.puzzle->setReadOnly(true);
		d->widgets.puzzle->setLetrs(d->game->puzzle().c_str());
		d->widgets.puzzle->setReadOnly(true);

		d->dsp->init(d->game->words(), wpl_calc());

		if (d->settings.savedGame) {
			switch (h.hint) {
			case 2:
				d->dsp->setHint();
				/* FALLTHROUGH */
			case 1:
				d->dsp->setHint();
				/* FALLTHROUGH */
			default:
				break;
			}
			enableHints(2 - h.hint);
			for (int i = 0; i < h.nwords; i++) {
				d->dsp->setWord(setWords[i]);
			}
		}

		d->widgets.wordsOutput->setHtml(d->dsp->html());
		d->widgets.statusWordCnt->setText(QString("Words:  %1 / %2 ")
										  .arg(d->dsp->nSetWords())
										  .arg(d->dsp->nWords()));
		d->t = d->settings.savedGame ? h.time : 0;
		d->timer->start();
		d->widgets.statusMsg->clear();
		if (d->settings.showTime)
			updateTime();
		if (d->settings.savedGame)
			d->widgets.statusOutput->showMessage("Game loaded.", 3000);
	} else {
		if (d->settings.savedGame)
			QMessageBox::information(this, "Game", "Failed to load saved game.");
		else
			QMessageBox::information(this, "Game", "No suitable game found. "
			       	 "Choose a different Solution range or use "
					 "a different dictionary.");

		d->widgets.puzzle->setReadOnly(false);
		d->widgets.solve->setEnabled(true);
		enableHints(-1);
		d->widgets.finish->setEnabled(false);
		d->widgets.wordInput->setEnabled(false);
		d->widgets.wordInput->wordEdit->setInitState(false);
		d->widgets.puzzle->clear();
		d->widgets.puzzle->setReadOnly(false);
		d->widgets.wordsOutput->clear();
		d->widgets.statusWordCnt->setText("          ");
		d->widgets.statusMsg->clear();
	}

	d->settings.savedGame = false;
	d->actions.settings->setEnabled(true);
	d->widgets.play->setEnabled(true);
}

void 
MainWindow::finish()
{
	d->widgets.statusOutput->clearMessage();
	d->widgets.play->setEnabled(true);
	enableHints(-1);
	d->widgets.finish->setEnabled(false);
	d->widgets.solve->setEnabled(true);
	d->widgets.play->setFocus();
	d->widgets.wordInput->wordEdit->chk_clear();
	d->widgets.wordInput->setEnabled(false);
	d->widgets.puzzle->setReadOnly(false);

	d->timer->stop();

	QString time("                ");

	if (d->settings.showTime) 
		time = QString("Time: &nbsp; %1:%2      ")
						.arg(QString::number((int)d->t/60), 2, QChar('0'))
						.arg(QString::number((int)d->t%60), 2, QChar('0'));

	d->widgets.statusWordCnt->setText("          ");
	d->widgets.statusMsg->clear();

	if (d->settings.showScore) {
		d->game->calcScore(d->dsp->nSetWords(), d->dsp->nWords(), 
			 	 d->dsp->hintLevel(), d->t, d->dsp->entered7());

		string rating = (d->game->rating() == "") ? "" : "[ <i><b><font color=\"" + d->game->rating_color() + "\">" + d->game->rating() + "</font></b></i> ]";
		QMessageBox m(QMessageBox::NoIcon, "Score", 
					  QString("<table cellpadding=\"12\" width=\"100\">"
							  "<tr><td>Correct words: &nbsp;%1 / %2</td><td>%5</td></tr>"
						      "<tr><td>Score: &nbsp; %3 &nbsp; %4</td></tr></table>")
						      .arg(d->dsp->nSetWords()).arg(d->dsp->nWords())
						      .arg(d->game->score()).arg(rating.c_str())
							  .arg(time)
							  , QMessageBox::Ok, this);
		m.setTextFormat(Qt::RichText);
		m.setIconPixmap(QIcon(":/spth.ico").pixmap(32, 32));
		m.exec();
	} 

	d->dsp->setSolved();
	d->widgets.wordsOutput->setHtml(d->dsp->html());

	d->state.ingame = false;
	d->state.inprogress = false;
}

void
MainWindow::solve()
{
	d->widgets.statusOutput->clearMessage();
	d->widgets.wordsOutput->clear();
	d->widgets.wordInput->setEnabled(false);
	d->widgets.statusWordCnt->setText("          ");
	d->widgets.statusMsg->clear();

	if (d->state.dictChanged != "") {
		if (!initBackend())
			return;
		d->widgets.statusOutput->showMessage("New dictionary loaded.", 3000);
		d->state.dictChanged = "";
	}

	QString s = d->widgets.puzzle->letrs();
	if (s != "") {
		int n = d->game->solve(s.toStdString());
		if (n > 0) {
			d->dsp->init(d->game->words(), wpl_calc());
			d->dsp->setSolved();

			d->widgets.wordsOutput->setHtml(d->dsp->html());
			d->widgets.statusWordCnt->setText(QString("Words:  %1 ")
						.arg(d->dsp->nWords()));
		} else {
			d->widgets.statusWordCnt->setText("Words:  0 ");
			d->widgets.statusMsg->setText("No Solution for given puzzle.");
		}

		d->widgets.play->setFocus();
	} else {
		d->widgets.puzzle->clear();
		QMessageBox::information(this, "Solve", 
					 "Please enter the seven letters of the puzzle in the"
					 " corresponding boxes and then press Solve.");
		d->widgets.puzzle->setReadOnly(false);
		d->widgets.puzzle->centralLetr()->setFocus();
	}
}

void 
MainWindow::wordEntered()
{
	switch (d->dsp->setWord(d->widgets.wordInput->wordEdit->text().toStdString())) {
	case 1:
		d->state.inprogress = true;

		d->widgets.statusWordCnt->setText(QString("Words:  %1 / %2 ")
				  .arg(d->dsp->nSetWords())
				  .arg(d->dsp->nWords()));
		d->widgets.wordsOutput->setHtml(d->dsp->html());
		d->widgets.wordInput->wordEdit->chk(WordInput::CORRECT);

		if (d->dsp->solved()) 
			finish();
		
		break;

	case -1:
		d->widgets.wordsOutput->setHtml(d->dsp->html());
		d->widgets.wordInput->wordEdit->chk(WordInput::WRONG);
		break;

	case 0:
		d->widgets.wordsOutput->setHtml(d->dsp->html());
		d->widgets.wordInput->wordEdit->chk(WordInput::REPEAT);
		break;
	}
}

void
MainWindow::resizeEvent(QResizeEvent *event)
{
	wpl_update();
	event->accept();
}

void
MainWindow::setStateFlPath()
{
#ifdef WIN32
		strcpy(stateFlPath, getenv("APPDATA"));
		strcat(stateFlPath, "/"GAMESTATEFL);
#else
		struct passwd *pw = getpwuid(getuid());
		if (pw == 0)
			Spellathon::spth_error(1, "Cannot get user info.");

		strcpy(stateFlPath, pw->pw_dir);
		strcat(stateFlPath, "/.spellathon/");
		if (access(stateFlPath, R_OK | W_OK) < 0) {
			if (mkdir(stateFlPath, 
			          S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
 			       Spellathon::spth_error(1, "Cannot create '%s'", stateFlPath);
		}
			
		strcat(stateFlPath, GAMESTATEFL);
#endif /* WIN32 */
}

void
MainWindow::closeEvent(QCloseEvent *event)
{
	d->settings.mainwindowSize = size();
	d->settings.mainwindowPos = pos();

	saveGame();
	d->settings.savedGame = d->state.ingame;
	event->accept();
}

MainWindow::~MainWindow()
{
	delete d; 
}

void 
MainWindow::wpl_update()
{
	d->dsp->setWpl(wpl_calc());
	d->widgets.wordsOutput->setText(d->dsp->html());
}

int
MainWindow::wpl_calc()
{
	int sz = d->widgets.wordsOutput->viewport()->width();
	int wsz = d->widgets.wordsOutput->fontMetrics()
										.width(QString(d->dsp->fw(), 'z'));
	return sz/wsz;
}

void
MainWindow::updateTime()
{
	++d->t;
	if (d->settings.showTime)
		d->widgets.statusMsg->setText(QString("Time:  %1:%2")
				.arg(QString::number((int)d->t/60), 2, QChar('0'))
				.arg(QString::number((int)d->t%60), 2, QChar('0')));
	else
		d->widgets.statusMsg->clear();
}
	
void
MainWindow::hint()
{
	d->dsp->setHint();
	d->widgets.wordsOutput->setHtml(d->dsp->html());

	switch (d->dsp->hintLevel()) {
	case 0:
			enableHints(2-0);
			break;
	case 1:
			enableHints(2-1);
			break;
	case 2:
	default:
			enableHints(2-2);
			d->widgets.statusOutput->showMessage("No more hints.", 3000);
			break;
	}
}

void
MainWindow::openWWW()
{
	QDesktopServices::openUrl(QUrl("http://sourceforge.net/projects/spellathon/"));
}
