/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <errno.h>

#ifndef WIN32
# include <sys/stat.h>
# include <pwd.h>
#endif /* WIN32 */

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>

#ifndef SPTH_CL
# include <QProgressDialog>
# include <QFileInfo>
# include <QDateTime>
#include <QMessageBox>
#endif /* SPTH_CL */

#ifdef WIN32
# include <QTime>
#else
# include <sys/time.h>
#endif /* WIN32 */

#include "dict.h"
#include "util.h"
#include "game.h"

/* 
	The following macros might bring tears to your eyes :D
	I can't think of a better way to avoid code duplication.
	Don't know how to use templates in this case or if its even possible.
*/

#define IDXOFF(idxarr, sign, len, indx, end)								\
		(idxarr[(len)-indx][(sign)[0]-'a'][(sign)[1]-'a'][(sign)[2]-'a'].end)

#define bsearch7(l, h, compr)											\
	do {																\
		int	m, r;														\
																		\
		if (l + h >= 0) { 												\
			while (h >= l) {											\
				m = (l + h) / 2;										\
				r = (compr);											\
				if (r < 0) 												\
					h = m - 1;											\
				else if (r > 0)											\
					l = m + 1;											\
				else													\
					return m;											\
			}															\
		}																\
																		\
		return -1;														\
	} while (0);

#define buildIndex(list, sign, len_pfx, indx)							\
	do {																\
		memset(list.idxOff, -1, sizeof(list.idxOff));					\
		memset(list.lenOff, -1, sizeof(list.lenOff));					\
		int curlen = 0;													\
		char cursignidx[3] = { 0, 0, 0 };								\
		for (int i = 0; i < list.size; i++) {							\
			int len = list.arr[i]len_pfx;								\
			if (len > curlen) {											\
				list.lenOff[len-indx].start = i;						\
				curlen = len;											\
				cursignidx[0] = cursignidx[1]= cursignidx[2]= 0;		\
			}															\
			list.lenOff[len-indx].end = i;								\
			if (list.arr[i]sign[0] != cursignidx[0] ||					\
				list.arr[i]sign[1] != cursignidx[1] ||					\
				list.arr[i]sign[2] != cursignidx[2] )					\
			{															\
				cursignidx[0] = list.arr[i]sign[0];						\
				cursignidx[1] = list.arr[i]sign[1];						\
				cursignidx[2] = list.arr[i]sign[2];						\
				IDXOFF(list.idxOff, cursignidx, curlen, indx, start)=i;	\
			}															\
			IDXOFF(list.idxOff, cursignidx, curlen, indx, end) = i;		\
		}																\
	} while (0);

namespace Spellathon
{
	static int dictSortcmp(const void *p1, const void *p2);
	static int precompSortcmp(const void *p1, const void *p2);
	static int wordSortcmp(const void *p1, const void *p2);
	static int nonPluralSortcmp(const void *p1, const void *p2);

	void
	Dict::dictOpen(const char *path)
	{
		char	line[MAX_LNLEN], *r; 
		int		slen;
		FILE	*fd;
	
		fd = fopen(path, "r");
		if (fd == 0) 
			spth_error(1, "Cannot open dictionary: '%s'", path);
	
		wordLst_.size = 0;
		int sz = 1<<14;
	
		wordLst_.arr = (WordEnt*)memalloc(sz * sizeof(WordEnt));
		memset(wordLst_.arr, 0, sz * sizeof(WordEnt));
	
again:
		while ( (r = fgets(line, MAX_LNLEN, fd)) != NULL) {
			slen = strlen(line);

			if (line[slen-1] != '\n') {
				fclose(fd);
				destroy();
				errno = EINVAL;
				spth_error(0, "'%s': Unrecognized file type. Dictionary must "
						   "be an ascii text file with the words seperated by newlines."
						   , path);
			}

			/* Remove newline char */
			line[--slen] = 0;

			/* Remove /r if file is in Windows format */
			if (line[slen-1] == '\r')
				line[--slen] = 0;
	
			if (slen < MIN_LETRS-1 || slen > MAX_LETRS)
				goto again;
	
			for (int i = 0; i < slen; i++) 
				if(!islower(line[i]))
					goto again;

			if (wordLst_.size >= sz) {
				wordLst_.arr = (WordEnt*)memresize(wordLst_.arr, sz * sizeof(WordEnt) * 2);
				memset(wordLst_.arr + sz, 0, sz * sizeof(WordEnt));
				sz *= 2;
			}
			strcpy(wordLst_.arr[wordLst_.size].word, line);
			wordLst_.arr[wordLst_.size++].len = slen;
		}
	
		wordLst_.arr = (WordEnt*)memresize(wordLst_.arr, wordLst_.size * sizeof(WordEnt));

		if (ferror(fd)) {
			fclose(fd);
			destroy();
			spth_error(1, "'%s': Read error", path);
		}
	
		fclose(fd);

		qsort((void *)wordLst_.arr, wordLst_.size, sizeof(WordEnt), wordSortcmp);

		wordBuildIndex();

		dictProcess();
		dictBuildIndex();

		if (pluralSieve_) {
			processNonPlural();
			nonPluralBuildIndex();
		}
	}

	void
	Dict::dictProcess()
	{
		int L3sz = wordLst_.lenOff[0].end + 1;
		dictLst_.size = wordLst_.size - L3sz;

		dictLst_.arr = (DictEnt *)memalloc(dictLst_.size * sizeof(DictEnt));
		memset(dictLst_.arr, 0, dictLst_.size * sizeof(DictEnt));

		char tword[MAX_LETRS+1];
		for (int i = 0; i <  dictLst_.size; i++) {
				dictLst_.arr[i].wordp = &wordLst_.arr[i+L3sz];
				strcpy7(tword, dictLst_.arr[i].wordp->word);
				sign(tword, dictLst_.arr[i].wordp->len);
				strcpy7(dictLst_.arr[i].sign, tword);
		}

		qsort((void*)dictLst_.arr, dictLst_.size, sizeof(DictEnt), 
			  dictSortcmp);

	}

	void
	Dict::processNonPlural()
	{
		int	s, e;
		int sz = 1<<10;
		nonPluralLst_.size = 0;
		char word[MAX_LETRS+1];

		nonPluralLst_.arr = (WordEnt**)memalloc(sz * sizeof(WordEnt*));

		for (int i = wordLst_.lenOff[1].start; i < wordLst_.size; i++) {
			strcpy7(word, wordLst_.arr[i].word);
			int slen = wordLst_.arr[i].len;
			/* .*s */
			if (word[--slen] == 's') {
				word[slen] = 0;	
				s = IDXOFF(wordLst_.idxOff, word, slen, 3, start);
				e = IDXOFF(wordLst_.idxOff, word, slen, 3, end);
				if (searchWord(word, s, e) != -1) 
					continue;
				/* .*es */
				if (word[--slen] == 'e') {
					word[slen] = 0;
					s = IDXOFF(wordLst_.idxOff, word, slen, 3, start);
					e = IDXOFF(wordLst_.idxOff, word, slen, 3, end);
					if (searchWord(word, s, e) != -1) 
						continue;
				}
				/* .*ies */
				if (word[--slen] == 'i') {
					word[slen] = 'y';
					s = IDXOFF(wordLst_.idxOff, word, slen+1, 3, start);
					e = IDXOFF(wordLst_.idxOff, word, slen+1, 3, end);
					if (searchWord(word, s, e) != -1) 
						continue;
				}
				if (nonPluralLst_.size >= sz) {
					sz *= 2;
					nonPluralLst_.arr = (WordEnt**)memresize(nonPluralLst_.arr, sz * sizeof(WordEnt*));
				}
				nonPluralLst_.arr[nonPluralLst_.size++] = &wordLst_.arr[i];
			}
		}

		qsort((void *)nonPluralLst_.arr, nonPluralLst_.size, 
			  sizeof(WordEnt*), nonPluralSortcmp);

	}

	void
	Dict::wordBuildIndex()
	{
		buildIndex(wordLst_, .word, .len, 3);
	}

	void
	Dict::dictBuildIndex()
	{
		buildIndex(dictLst_, .sign, .wordp->len, 4);
	}

	void
	Dict::nonPluralBuildIndex()
	{
		buildIndex(nonPluralLst_, ->word, ->len, 4);
	}

#ifdef SPTH_CL
	Dict::Dict() 
#else
	Dict::Dict(QWidget* p) 
		: gui_h_(p)
#endif /* SPTH_CL */
	{ 
		dictbinPath_[0] = '\0';
		dictLst_.arr = 0;
		wordLst_.arr = 0;
		precompLst_.arr = 0;
		nonPluralLst_.arr = 0;
	}

	void
	Dict::init(const char *path, bool precomp, bool pluralSieve) 
	{
		dictPath_ = path;
		destroy();
		precomp_ = precomp;
		pluralSieve_ = pluralSieve;
		if (precomp_) {
			setPrecompPath();
			precompOpen(path); 
		} else {
			dictOpen(path); 
		}
	}
	
	void
	Dict::setPrecompPath()
	{
#ifdef WIN32
		strcpy(dictbinPath_, getenv("APPDATA"));
		strcat(dictbinPath_, "/"DICTBIN);
#else
		struct passwd *pw = getpwuid(getuid());
		if (pw == 0)
			spth_error(1, "Cannot get user info.");

		strcpy(dictbinPath_, pw->pw_dir);
		strcat(dictbinPath_, "/.spellathon/");
		if (access(dictbinPath_, R_OK | W_OK) < 0) {
			if (mkdir(dictbinPath_, 
			          S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
 			       spth_error(1, "Cannot create '%s'", dictbinPath_);
		}
			
		strcat(dictbinPath_, DICTBIN);
#endif /* WIN32 */
	}

	void
	Dict::precompOpen(const char *path)
	{
		FILE *binfl = fopen(dictbinPath_, "rb");	
		if (binfl == 0) {
			if (errno == ENOENT) {
				precompCreate(path);
				destroy();
				precompOpen(path);
				return;
			} else {
 			       spth_error(1, "Cannot open precomputed dictionary: '%s'", dictbinPath_);
			}
		}
			

		if (fread(&precompHdr_, sizeof(precompHdr_), 1, binfl) < 
			(unsigned int)1) {
			goto oops;
		}

		/* 
		 * perform sanity checks 
		 */
#ifdef SPTH_CL 
# ifndef WIN32
		struct stat binp, dictp;
		stat(dictbinPath_, &binp);
		stat(path, &dictp);
# endif
#endif /* SPTH_CL */
		if (
			cksum16(&precompHdr_, sizeof(precompHdr_)) != 0 ||
			strcmp(precompHdr_.dictPath, path) ||
		    strcmp(precompHdr_.magic, MAGIC) ||
			precompHdr_.version != PRECOMP_VER ||
			(precompHdr_.pluralSieve > 0) != pluralSieve_
#ifdef SPTH_CL 
# ifndef WIN32
			|| binp.st_size != precompHdr_.fsize
		    || binp.st_mtime < dictp.st_mtime
# endif
#else
			|| QFileInfo(dictbinPath_).size() != (int64_t)precompHdr_.fsize
		    || QFileInfo(dictbinPath_).lastModified().toTime_t() <
		    QFileInfo(path).lastModified().toTime_t()
#endif /* SPTH_CL */
		) {
			fclose(binfl);
			precompCreate(path);
			destroy();
			precompOpen(path);
			return;
		}

		wordLst_.size = precompHdr_.wordLst_size;
		wordLst_.arr = (WordEnt*)memalloc(wordLst_.size * sizeof(WordEnt));
		if (fread(wordLst_.arr, sizeof(WordEnt), wordLst_.size, binfl) < 	
			(unsigned int)wordLst_.size) 
			goto oops;

		precompLst_.size = precompHdr_.precompLst_size;
		precompLst_.arr = (PrecompEnt*)memalloc(precompLst_.size * sizeof(PrecompEnt));
		if (fread(precompLst_.arr, sizeof(PrecompEnt), 
					precompHdr_.precompLst_size, binfl) < 
			(unsigned int)precompHdr_.precompLst_size) 
			goto oops;

		wordBuildIndex();

		dictProcess();
		dictBuildIndex();

		if (pluralSieve_) {
			processNonPlural();
			nonPluralBuildIndex();
		}

		fclose(binfl);
		return;

oops:
		fclose(binfl);
		destroy();
		spth_error(1, "'%s': Read error", dictbinPath_);
	}

	void
	Dict::precompCreate(const char *path)
	{
		int cnt = 0, n;

		dictOpen(path);
		Game	game(this, pluralSieve_);

		precompLst_.arr = (PrecompEnt*)memalloc(nent7() * 7 * sizeof(PrecompEnt));

#ifndef SPTH_CL
		QProgressDialog *pdiag =  new QProgressDialog("Precomputing Dictionary ...", 0, 
							  dictLst_.lenOff[7-4].start , dictLst_.lenOff[7-4].end, gui_h_);
		pdiag->setWindowModality(Qt::WindowModal);
		pdiag->setMinimumDuration(0);
		pdiag->setWindowFlags(Qt::Popup);
		pdiag->move((gui_h_->size().width()-pdiag->size().width())/2 + gui_h_->pos().x(),
				   gui_h_->size().height()/2 + gui_h_->pos().y());
#endif /* SPTH_CL */

#ifdef WIN32
		QTime t;
		t.start();
#else
		struct timeval s, e;
		gettimeofday(&s, NULL);
#endif /* WIN32 */

		if (dictLst_.lenOff[7-4].start != -1) {
			for (int i = dictLst_.lenOff[7-4].start; 
				 i <= dictLst_.lenOff[7-4].end; i++) {
				char sig[MAX_LETRS+1];
				strcpy7(sig, dictLst_.arr[i].sign);
	
				char cltried[MAX_LETRS];
				int k = MAX_LETRS - 2;
				int clidx = 0;
				memset(cltried, 0, sizeof(cltried));
				do {
					if ((n = game.solver(sig)) > 0) {
						strcpy7(precompLst_.arr[cnt].sign, sig);	
						precompLst_.arr[cnt++].nsol = n;
					}

					/* find non-duplicate central letter */
					cltried[clidx++] = sig[MAX_LETRS-1];
again:
					for (char *p = cltried; *p && k >= 0; p++) {
						if (*p == sig[k]) {
							k--;
							goto again;
						} 
					}

					if (k < 0)
						break;

					/* try a different central letter */	
					swapch(&sig[MAX_LETRS-1], &sig[k--]);
				} while (1);

				if (!(i & ((1<<9) - 1))) {
#ifndef SPTH_CL
					pdiag->setValue(i);
#else
					printf("\rPrecomputing dictionary... %d%%", (i-dictLst_.lenOff[7-4].start)*100/nent7());
					fflush(stdout);
#endif /* SPTH_CL */
				}
			}
		}

		/* sort by number of solutions */
		qsort((void *)precompLst_.arr, cnt, sizeof(PrecompEnt), precompSortcmp);

		/* remove dup signs */
		if (cnt) {
			PrecompEnt *f = precompLst_.arr;
			PrecompEnt *l = precompLst_.arr + cnt;
			PrecompEnt *r = f;
			while (++f != l) {
				if (strcmp(r->sign, f->sign))
					memmove(++r, f, sizeof(PrecompEnt));
			}
			++r;
			cnt = r - precompLst_.arr;
		}

		/* fill header */
		memset(&precompHdr_, 0, sizeof(precompHdr_));
		precompHdr_.cksum = 0;
		precompHdr_.wordLst_size = wordLst_.size;
		precompHdr_.precompLst_size = cnt;
		precompHdr_.version = PRECOMP_VER;
		strncpy(precompHdr_.dictPath, path, MAX_PATH_LEN);
		precompHdr_.dictPath[MAX_PATH_LEN-1] = 0;
		strcpy(precompHdr_.magic, MAGIC);
		precompHdr_.pluralSieve = pluralSieve_;
		precompHdr_.fsize = sizeof(precompHdr_) + 
				sizeof(WordEnt)*wordLst_.size + 
				sizeof(PrecompEnt)*cnt;
		
		/* internet 16bit cksum for the header */
		precompHdr_.cksum = cksum16(&precompHdr_, sizeof(precompHdr_));

		FILE *binfl = fopen(dictbinPath_, "wb");
		if (binfl == 0) {
			destroy();
			spth_error(1, "Cannot open precomputed dictionary: '%s'", dictbinPath_);
		}

		if (
			fwrite(&precompHdr_, sizeof(precompHdr_), 1, binfl) < 1 ||
			fwrite(wordLst_.arr, sizeof(WordEnt), wordLst_.size, binfl) < 
				(unsigned int)wordLst_.size ||
			fwrite(precompLst_.arr, sizeof(PrecompEnt), cnt, binfl) < 
				(unsigned int)cnt
		   ) {
			fclose(binfl);
			destroy();
			spth_error(1, "'%s': Write error", dictbinPath_);
		}

#ifdef WIN32
		double et = (double)t.elapsed()/1000;
#else
		gettimeofday(&e, NULL);
		double et = e.tv_sec - s.tv_sec + (double)(e.tv_usec - s.tv_usec)/1000000;
#endif /* WIN32 */

#ifdef SPTH_CL
		printf("\rDictionary precomputed in %.3f secs.\n"
			   "Press any key to continue...", et);
		getchar();
#else
			delete pdiag;
			QMessageBox::information(gui_h_, "Precomputing dictionary...", QString("Dictionary precomputed in %1 secs.").arg(et, 0, 'f', 3));
#endif /* SPTH_CL */

		fclose(binfl);
	}

	/* binary search the dict. */
	/* performance critical routine */
	inline int
	Dict::searchSign(const char * sign, int l, int h) const
	{
		bsearch7(l, h, strcmp7(sign, dictLst_.arr[m].sign));
	}

	inline int
	Dict::searchNonPlural(const char *word, int l, int h) const
	{
		bsearch7(l, h, strcmp7(word, nonPluralLst_.arr[m]->word));
	}

	int
	Dict::searchWord(const char *word, int l, int h) const
	{
		bsearch7(l, h, strcmp7(word, wordLst_.arr[m].word));
	}
	
	
	void
	Dict::matchSign(const char * sign, char *w[], int *idx, int *has7) const
	{
		int pos, npos, slen, nw = 0;
	
		slen = strlen(sign);
	
		int s = IDXOFF(dictLst_.idxOff, sign, slen, 4, start);
		int e = IDXOFF(dictLst_.idxOff, sign, slen, 4, end);

		/* binary search dictionary for matching signature. */
		pos = searchSign(sign, s, e);
	
		if (pos != -1) {
			/* 
			 * search for all remaining duplicate signatures.
			 * the duplicates can be in contiguous locations
			 * on both the left and right sides of the index 
			 * returned by searchSign(). 
			 */
			npos = pos - 1;
			/* scan left */
			while (npos >= s && strcmp7_eq(dictLst_.arr[npos].sign, sign)) {
				char *p = dictLst_.arr[npos].wordp->word;
				if (
					pluralSieve_ && 
					p[slen-1] == 's' &&
					searchNonPlural(
							 p, 
							 IDXOFF(nonPluralLst_.idxOff, p, slen, 4, start), 
							 IDXOFF(nonPluralLst_.idxOff, p, slen, 4, end)
							) == -1
				   )
				{
					npos--;
					continue;
				}

				w[(*idx)++] = p;
				nw++;

				if (*idx > SOLVER_MAX_WORDS)
					return;

				npos--;
			}
			/* scan right */
			do {
				char *p = dictLst_.arr[pos].wordp->word;
				if (
					pluralSieve_ && 
					p[slen-1] == 's' &&
					searchNonPlural(
							 p, 
							 IDXOFF(nonPluralLst_.idxOff, p, slen, 4, start), 
							 IDXOFF(nonPluralLst_.idxOff, p, slen, 4, end)
							) == -1
				   )
				{
					continue;
				}

				w[(*idx)++] = p;
				nw++;

				if (*idx > SOLVER_MAX_WORDS)
					return;
			} while (++pos <= e && strcmp7_eq(dictLst_.arr[pos].sign, sign));
		}

		if (slen == 7)
			*has7 = nw;
	}
	
	/* sort by strlen() and then lexicographically */
	static int
	dictSortcmp(const void *p1, const void *p2)
	{
		const DictEnt *s1 = (const DictEnt*)p1;
		const DictEnt *s2 = (const DictEnt*)p2;
	
		int r = s1->wordp->len - s2->wordp->len;
		if (r == 0) 
			return strcmp7(s1->sign, s2->sign);
		else 
			return r;
	}
	
	static int
	wordSortcmp(const void *p1, const void *p2)
	{
		const WordEnt *s1 = (const WordEnt*)p1;
		const WordEnt *s2 = (const WordEnt*)p2;
		
		int r = s1->len - s2->len;
		if (r == 0) 
			return strcmp7(s1->word, s2->word);
		else 
			return r;
	}
	
	static int
	precompSortcmp(const void *p1, const void *p2)
	{
		const PrecompEnt *s1 = (const PrecompEnt *)p1;
		const PrecompEnt *s2 = (const PrecompEnt *)p2;
		
		int r = s1->nsol - s2->nsol;
		if (r == 0) 
			return strcmp7(s1->sign, s2->sign);
		else 
			return r;
	}

	static int
	nonPluralSortcmp(const void *p1, const void *p2)
	{
		const WordEnt *s1 = *(const WordEnt **)p1;
		const WordEnt *s2 = *(const WordEnt **)p2;
		
		int r = s1->len - s2->len;
		if (r == 0) 
			return strcmp7(s1->word, s2->word);
		else 
			return r;
	}
	
	void
	Dict::destroy()
	{
		free(dictLst_.arr);			dictLst_.arr = 0;
		free(wordLst_.arr);			wordLst_.arr = 0;
		free(nonPluralLst_.arr); 	nonPluralLst_.arr = 0;
		free(precompLst_.arr);		precompLst_.arr = 0;
	}
	
	uint16_t
	Dict::cksum16(void *addr, int bytes)
	{
		unsigned int sum = 0;
		uint16_t *p = (uint16_t*)addr;

		while (bytes > 1) {
			sum += *p++;
			bytes -= 2;
		}
	
		if (bytes > 0)
			sum += *p;

		while (sum >> 16)
			sum = (sum & 0xffff) + (sum >> 16);

		return ~sum;
	}
}
